#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <tgmath.h>
#include <mpi.h>

/*
 * Physical quantities:
 * k                    : thermal conductivity      [Watt / (meter Kelvin)]
 * rho                  : density                   [kg / meter^3]
 * cp                   : specific heat capacity    [kJ / (kg Kelvin)]
 * rho * cp             : volumetric heat capacity  [Joule / (meter^3 Kelvin)]
 * alpha = k / (rho*cp) : thermal diffusivity       [meter^2 / second]
 *
 * Mercury:
 * cp = 0.140, rho = 13506, k = 8.69
 * alpha = 8.69 / (0.140*13506) =~ 0.0619
 *
 * Copper:
 * cp = 0.385, rho = 8960, k = 401
 * alpha = 401.0 / (0.385 * 8960) =~ 0.12 [0.1162453618]
 *
 * Tin:
 * cp = 0.227, k = 67, rho = 7300
 * alpha = 67.0 / (0.227 * 7300) =~ 0.040
 *
 * Aluminium:
 * cp = 0.897, rho = 2700, k = 237
 * alpha = 237 / (0.897 * 2700) =~ 0.098 [0.097857054]
 */
#define MERCURY 0.0619
#define COPPER 0.116
#define TIN 0.040
#define ALUMINIUM 0.098


/* Size of the computational grid - 256x256 square */
#define SIZE 256 //was 256    

/* Write to File 1=true and 0=false*/
#define WRITETOFILE 0

/* Parameters of the simulation: how many steps, and when to cut off the heat */
#define NSTEPS 125000 //was 125000
#define CUTOFF 75000 //was 75000

/* How often to dump state to file (steps).
 * 16 is realtime at 25fps, this is in 10x time
 */
#define SNAPSHOT 160 //was 160

/* Local material constant (LMAT) and temperature (LTEMP) indexing macros */
#define MAT(i,j)   material[                             \
        ((i)*systemSIZE)+(j)        \
]

#define TEMP(s,i,j)    temperature[((s)%2)][((i)*systemSIZE)+(j)]

/* Arrays for the simulation data */
float
    *temperature[2],       // Temperature field (in global domain on rank 0)
    *material;             // Local part of the material constants

/* Variables for time measurement */
double start, end;

int systemSIZE = SIZE;

/* Discretization: 5cm square cells, 2.5ms time intervals */
const float
    h  = 5e-2,
    dt = 2.5e-3;

void logTime(void);
void ftcs_solver ( int step );
void external_heat ( int step );
void configure_geometry ( void );
void collect_area ( int step, char *filename );
void write_matrix ( FILE *out, float *data );


void
ftcs_solver ( int step )
{
    /* my and mx are the largest y and x numbers in the main area of the local_temp and local_material matrices */
    int my = systemSIZE-1, mx = systemSIZE-1;

    /* The FTCS solution */
    for ( int y=1; y<my; y++ ){
        for ( int x=1; x<mx; x++ ){
            TEMP(step+1,y,x) = TEMP(step,y,x) + MAT(y,x) * (
                (TEMP(step,y-1,x) + TEMP(step,y+1,x) +
                 TEMP(step,y,x-1) + TEMP(step,y,x+1)) - 4.0*TEMP(step,y,x)
            );
        }
    }
    for ( int i=1; i<my; i++ )
        TEMP(step+1,i,0) = TEMP(step,i,0) + MAT(i,0) * (
            (2*TEMP(step,i,1) + TEMP(step,i-1,0) + TEMP(step,i+1,0))
            - 4.0*TEMP(step,i,0)
        );
    for ( int i=1; i<my; i++ )
            TEMP(step+1,i,mx) = TEMP(step,i,mx) + MAT(i,mx) * (
                (2*TEMP(step,i,mx-1) + TEMP(step,i-1,mx) + TEMP(step,i+1,mx))
                - 4.0*TEMP(step,i,mx)
            );
    for ( int i=1; i<mx; i++ )
        TEMP(step+1,0,i) = TEMP(step,0,i) + MAT(0,i) * (
            2*TEMP(step,1,i) + TEMP(step,0,i-1) + TEMP(step,0,i+1)
            - 4.0 * TEMP(step,0,i)
        );
    for ( int i=1; i<mx; i++ )
        TEMP(step+1,my,i) = TEMP(step,my,i) + MAT(my,i) * (
            2*TEMP(step,my-1,i) + TEMP(step,my,i-1) + TEMP(step,my,i+1)
            - 4.0 * TEMP(step,my,i)
        );

    TEMP(step+1,0,0) = TEMP(step,0,0) + MAT(0,0) * (
        2*TEMP(step,1,0) + 2*TEMP(step,0,1) - 4.0 * TEMP(step,0,0)
    );
    TEMP(step+1,0,mx) = TEMP(step,0,mx) + MAT(0,mx) * (
            2*TEMP(step,1,mx) + 2*TEMP(step,0,mx-1) - 4.0 * TEMP(step,0,mx)
    );
    TEMP(step+1,my,0) = TEMP(step,my,0) + MAT(my,0) * (
        2*TEMP(step,my,1) + 2*TEMP(step,my-1,0) - 4.0 * TEMP(step,my,0)
    );
    TEMP(step+1,my,mx) = TEMP(step,my,mx) + MAT(my,mx) * (
        2*TEMP(step,my-1,mx)+2*TEMP(step,my,mx-1) - 4.0*TEMP(step,my,mx)
    );
}

void logTime( void){
        char* filename= "log.txt";
        FILE *out = fopen ( filename, "a" );
        if (out == NULL)
        {
            printf("Error opening file!\n");
            exit(1);
        }
        fprintf ( out, "Runtime = %f systemSIZE %d NSTEPS %d CUTOFF%d THREADS %d in y=%d,x=%d WriteTofile(%d)\n", end-start, systemSIZE, NSTEPS, CUTOFF, 1, 1, 1, WRITETOFILE);
        fclose ( out );

}

void printLargeTemperature(int step){
    	for ( int y=0; y<systemSIZE; y++ ){
        	for ( int x=0; x<systemSIZE; x++ )
        	{
                printf(" %3.1f ",temperature[(step%2)][y*systemSIZE + x]);
        	}
        	printf("\n");
    	}
    
}

int
main ( int argc, char **argv )
{
    MPI_Init ( &argc, &argv );
    if(argc > 1){
        int argv1 = atoi(argv[1]);
        if((argv1%256) == 0){
            systemSIZE = argv1;
        }
    }
    int rank;
    MPI_Comm_rank( MPI_COMM_WORLD, &rank ) ;
if(rank == 0){
    start = MPI_Wtime();
    material = malloc ( systemSIZE * systemSIZE * sizeof(float) );
    temperature[0] = malloc(systemSIZE*systemSIZE * sizeof(float));
    temperature[1] = malloc(systemSIZE*systemSIZE * sizeof(float));

    configure_geometry(); //Set up the LMAT and LTEMP

    /* Main integration loop: NSTEPS iterations, impose external heat
     * until CUTOFF iterations have passed
     */
    /* Imposed temperature from outside */
    for ( int step=0; step<NSTEPS; step++ )
    {
        if ( step < CUTOFF )
            external_heat ( step );
        ftcs_solver ( step );
        if((step % SNAPSHOT) == 0 )
        {
            char filename[15];
            sprintf ( filename, "data/%.4d.dat", step/SNAPSHOT );
            collect_area ( step, filename );
        }
    }

    /* End timing */
    end = MPI_Wtime();

    free (temperature[0]);
    free (temperature[1]);
    free (material);

    printf("RUNTIME %4.6f systemSIZE %d NSTEPS %d CUTOFF%d THREADS %d in y %d x %d\n", end-start, systemSIZE, NSTEPS, CUTOFF, 1, 1, 1);
    logTime();
}
    MPI_Finalize();
    exit ( EXIT_SUCCESS );
}


void
external_heat ( int step )
{
    /* Imposed temperature from outside */
    for ( int y=(systemSIZE/2)-(systemSIZE/16); y<=(systemSIZE/2)+(systemSIZE/16); y++ )
        for ( int x=(systemSIZE/4); x<=(3*systemSIZE/4); x++ )
        {
            TEMP ( step, y, x) = 100.0;
        }
}


void
configure_geometry ( void )
{
    /* Initialization: fill the pool with mercury */
    for ( int y=0; y<systemSIZE; y++ )
    {
        for ( int x=0; x<systemSIZE; x++ )
        {
            MAT(y,x) =  MERCURY * (dt/(h*h));
            TEMP(1,y,x) = TEMP(0,y,x) = 20.0;
        }
    }

    /* Set up the two blocks of copper and tin */
    for ( int y=(systemSIZE/8); y<(3*systemSIZE/8); y++ )
        for ( int x=(systemSIZE/8); x<(systemSIZE/2)-(systemSIZE/8); x++ )
        {
            MAT(y,x) = COPPER * (dt/(h*h));
            TEMP(0,y,x) = 60.0;
            
            MAT(y,(systemSIZE-x)) = TIN * (dt/(h*h));
            TEMP(0,y,(systemSIZE-x)) = 60.0;
        }
    /* Set up the heating element in the middle */
    for ( int y=(systemSIZE/2)-(systemSIZE/16); y<=(systemSIZE/2)+(systemSIZE/16); y++ )
        for ( int x=(systemSIZE/4); x<=(3*systemSIZE/4); x++ )
        {
                MAT(y,x) = ALUMINIUM * (dt/(h*h));
        }
}


void
collect_area ( int step, char *filename )
{
    if((WRITETOFILE == 1)){
        //FILE *out = fopen ( filename, "w" );
        FILE *out = fopen ( "testwrite.dat", "w" );
        write_matrix ( out, temperature[((step)%2)] );
        fclose ( out );
        printf ( "Snapshot at step %d\n", step );
    }
}


void
write_matrix ( FILE *out, float *data )
{
    float size = (float) systemSIZE;
    fwrite ( &size, sizeof(float), 1, out );
    for ( float x=0; x<systemSIZE; x+=1.0 )
        fwrite ( &x, sizeof(float), 1, out );
    for ( int y=0; y<systemSIZE; y++ )
    {
        float len = (float) y;
        fwrite ( &len, sizeof(float), 1, out );
        fwrite ( &data[y*systemSIZE], sizeof(float), systemSIZE, out );
    }
}
