#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <tgmath.h>
#include <mpi.h>

/*
 * Physical quantities:
 * k                    : thermal conductivity      [Watt / (meter Kelvin)]
 * rho                  : density                   [kg / meter^3]
 * cp                   : specific heat capacity    [kJ / (kg Kelvin)]
 * rho * cp             : volumetric heat capacity  [Joule / (meter^3 Kelvin)]
 * alpha = k / (rho*cp) : thermal diffusivity       [meter^2 / second]
 *
 * Mercury:
 * cp = 0.140, rho = 13506, k = 8.69
 * alpha = 8.69 / (0.140*13506) =~ 0.0619
 *
 * Copper:
 * cp = 0.385, rho = 8960, k = 401
 * alpha = 401.0 / (0.385 * 8960) =~ 0.12 [0.1162453618]
 *
 * Tin:
 * cp = 0.227, k = 67, rho = 7300
 * alpha = 67.0 / (0.227 * 7300) =~ 0.040
 *
 * Aluminium:
 * cp = 0.897, rho = 2700, k = 237
 * alpha = 237 / (0.897 * 2700) =~ 0.098 [0.097857054]
 */
#define MERCURY 0.0619
#define COPPER 0.116
#define TIN 0.040
#define ALUMINIUM 0.098


/* Size of the computational grid - 256x256 square */
#define SIZE 256 //was 256

/* Write to File 1=true and 0=false*/
#define WRITETOFILE 0 

/* Parameters of the simulation: how many steps, and when to cut off the heat */
#define NSTEPS 125000 //was 125000
#define CUTOFF 75000 //was 75000

/* How often to dump state to file (steps).
 * 16 is realtime at 25fps, this is in 10x time
 */
#define SNAPSHOT 160 //was 160

/* Indexing macros for the global view on rank 0 */
#define TEMP(i,j) temperature[(i)*widthSIZE+(j)]

/* Test condition to see if a global coordinate is in my local area
 * BOX is within main area
 * BORBOX is whole local area
 */
#define BOX(y,x) (                                              \
     (y)>=local_origin[0] &&                                    \
     (y)<local_origin[0]+local_dims[0] &&                       \
     (x)>=local_origin[1] &&                                    \
     (x)<local_origin[1]+local_dims[1]                          \
)
#define BORBOX(y,x)(											\
	 (y)>=(local_origin[0]-border ) &&                           \
     (y)<(local_origin[0]+local_dims[0]+border) &&              \
     (x)>=(local_origin[1]- border ) &&                          \
     (x)<(local_origin[1]+local_dims[1]+border )  				\
)

/* Local material constant (LMAT) and temperature (LTEMP) indexing macros */
//#define border 1
#define LMAT(i,j)   local_material[                             \
        ((i)+border )*(local_dims[1]+2*border)+(j)+border        \
]

#define LTEMP(s,i,j)    local_temp[((s)%2)][                    \
    ((i)+border)*(local_dims[1]+2*border)+(j)+border            \
]

/* Arrays for the simulation data */
float
    *temperature,       // Temperature field (in global domain on rank 0)
    *local_material,    // Local part of the material constants
    *local_temp[2];     // Local part of the temperature (2 buffers)

/* Variables for time measurement */
double start, end;

/* Discretization: 5cm square cells, 2.5ms time intervals */
const float
    h  = 5e-2, //was 5e-2
    dt = 2.5e-3;

/* Local state */
int
    size, rank,                     // World size, my rank
    dims[2],                        // Size of the cartesian
    periods[2] = { false, false },  // Periodicity of the cartesian
    coords[2],                      // My coordinates in the cartesian
    north, south, east, west,       // Neighbors in the cartesian
    local_dims[2],                  // Size of local subdomain
    local_origin[2],                // World coordinates of (0,0) local
    local_realdims[2],				// Computing area
    padding[2] = {0,0},					// Size of padding added
    border = 1,
    systemSIZE = SIZE,
    widthSIZE = SIZE,
    heightSIZE = SIZE;

MPI_Comm cart;
MPI_Datatype
    global_area, local_area,        // Vectors for collecting subdomains
    border_row, border_col;         // Vectors for border exchange

void logTime( void );
void printStatus(int step); //TODO: delete
void printLargeTemperature( void ); //TODO: delete
void ftcs_solver ( int step );
void update_border( int step );
void boundaries ( int step );
void border_exchange ( int step );
void commit_vector_types ( void );
void external_heat ( int step );
void configure_geometry ( void );
void collect_area ( int step, char *filename );
void write_matrix ( FILE *out, float *data );


void
ftcs_solver ( int step )
{
    /* The FTCS solution */
    for ( int y=0; y<local_realdims[0]; y++ )
        for ( int x=0; x<local_realdims[1]; x++ )
            LTEMP(step+1,y,x) = LTEMP(step,y,x) + LMAT(y,x) * (
                (LTEMP(step,y-1,x) + LTEMP(step,y+1,x) +
                 LTEMP(step,y,x-1) + LTEMP(step,y,x+1)) - 4.0*LTEMP(step,y,x)
            );
}


void
boundaries ( int step )
{
    /* The Neumann boundary condition */

    /* my and mx are the largest y and x numbers in the main area of the local_temp and local_material matrices */
    int my = local_realdims[0]-1, mx = local_realdims[1]-1;

    // I have a west boundary
    if ( coords[1] == 0 ){
    	//Apply the stencil for the west column
        for ( int i=0; i<local_realdims[0]; i++ )
            LTEMP(step+1,i,0) = LTEMP(step,i,0) + LMAT(i,0) * (
                (2*LTEMP(step,i,1) + LTEMP(step,i-1,0) + LTEMP(step,i+1,0))
                - 4.0*LTEMP(step,i,0)
            );
    }

    // I have an east boundary
    if ( coords[1] == dims[1]-1 ){
    	//Apply the stencil for the east column
        for ( int i=0; i<local_realdims[0]; i++ )
            LTEMP(step+1,i,mx) = LTEMP(step,i,mx) + LMAT(i,mx) * (
                (2*LTEMP(step,i,mx-1) + LTEMP(step,i-1,mx) + LTEMP(step,i+1,mx))
                - 4.0*LTEMP(step,i,mx)
            );
    }

    // I have a north boundary
    if ( coords[0] == 0 ){
    	//Apply the stencil for the northern row
        for ( int i=0; i<local_realdims[1]; i++ )
            LTEMP(step+1,0,i) = LTEMP(step,0,i) + LMAT(0,i) * (
                2*LTEMP(step,1,i) + LTEMP(step,0,i-1) + LTEMP(step,0,i+1)
                - 4.0 * LTEMP(step,0,i)
            );
    }

    // I have a south boundary
    if ( coords[0] == dims[0]-1 ){
    	//Apply the stencil for the southern row
        for ( int i=0; i<local_realdims[1]; i++ )
            LTEMP(step+1,my,i) = LTEMP(step,my,i) + LMAT(my,i) * (
                2*LTEMP(step,my-1,i) + LTEMP(step,my,i-1) + LTEMP(step,my,i+1)
                - 4.0 * LTEMP(step,my,i)
            );
    }

    //Apply the stencil for the corners
    if ( BOX(0,0) )
        LTEMP(step+1,0,0) = LTEMP(step,0,0) + LMAT(0,0) * (
            2*LTEMP(step,1,0) + 2*LTEMP(step,0,1) - 4.0 * LTEMP(step,0,0)
        );
    if ( BOX(0,systemSIZE-1) )
        LTEMP(step+1,0,mx) = LTEMP(step,0,mx) + LMAT(0,mx) * (
            2*LTEMP(step,1,mx) + 2*LTEMP(step,0,mx-1) - 4.0 * LTEMP(step,0,mx)
        );
    if ( BOX(systemSIZE-1,0) )
        LTEMP(step+1,my,0) = LTEMP(step,my,0) + LMAT(my,0) * (
            2*LTEMP(step,my,1) + 2*LTEMP(step,my-1,0) - 4.0 * LTEMP(step,my,0)
        );
    if ( BOX(systemSIZE-1,systemSIZE-1) )
        LTEMP(step+1,my,mx) = LTEMP(step,my,mx) + LMAT(my,mx) * (
            2*LTEMP(step,my-1,mx)+2*LTEMP(step,my,mx-1) - 4.0*LTEMP(step,my,mx)
        );
}

void
update_border ( int step ){
	/* my and mx are the largest y and x numbers in the main area of the local_temp and local_material matrices */
    int my = local_realdims[0]-1, mx = local_realdims[1]-1;
    int startPos = (step%border )-border ;
	//Update westborder if there is a neighboor in west
	if(coords[1] != 0){
		//apply the stencil for the numbers in the border this must be done for the one colums closest to main area
		for(int x=startPos;x<0;x++){
			if(coords[0] == 0){
            	//can not use north border for y=0
            	LTEMP(step,0,x) = LTEMP(step-1,0,x) + LMAT(0,x) * (
            		(2.0*LTEMP(step-1,1,x) + LTEMP(step-1,0,x-1) +
            		LTEMP(step-1,0,x+1)) - 4.0*LTEMP(step-1,0,x)
            	);
            }else{
            	LTEMP(step,0,x) = LTEMP(step,0,x) = LTEMP(step-1,0,x) + LMAT(0,x) * (
                	(LTEMP(step-1,-1,x) + LTEMP(step-1,1,x) +
                 	LTEMP(step-1,0,x-1) + LTEMP(step-1,0,x+1)) - 4.0*LTEMP(step-1,0,x)
            	);
            }
            if(coords[0] == (dims[0]-1)){
            	//can not use south border for y=my since my+1 is wrong
            	LTEMP(step, my,x) = LTEMP(step-1,my,x) + LMAT(my,x) * (
                	(2.0*LTEMP(step-1,my-1,x) +
                 	LTEMP(step-1,my,x-1) + LTEMP(step-1,my,x+1)) - 4.0*LTEMP(step-1,my,x)
            	);
            }else{
            	LTEMP(step,my,x) = LTEMP(step-1,my,x) + LMAT(my,x) * (
                	(LTEMP(step-1,my-1,x) + LTEMP(step-1,my+1,x) +
                 	LTEMP(step-1,my,x-1) + LTEMP(step-1,my,x+1)) - 4.0*LTEMP(step-1,my,x)
            	);
            }
			for ( int y=1; y<my; y++ ){
            	LTEMP(step,y,x) = LTEMP(step-1,y,x) + LMAT(y,x) * (
                	(LTEMP(step-1,y-1,x) + LTEMP(step-1,y+1,x) +
                 	LTEMP(step-1,y,x-1) + LTEMP(step-1,y,x+1)) - 4.0*LTEMP(step-1,y,x)
            	);
            }
		}

	}


	//Update eastborder if there is a neighboor in east
	if(coords[1] != (dims[1]-1)){
		for(int x=(mx-startPos); x>mx; x--){
			if(coords[0] == 0){
            	//can not use north border for y=0
            	LTEMP(step,0,x) = LTEMP(step-1,0,x) + LMAT(0,x) * (
            		(2.0*LTEMP(step-1,1,x) + LTEMP(step-1,0,x-1) +
            		LTEMP(step-1,0,x+1)) - 4.0*LTEMP(step-1,0,x)
            	);
            }else{
            	LTEMP(step,0,x) = LTEMP(step,0,x) = LTEMP(step-1,0,x) + LMAT(0,x) * (
                	(LTEMP(step-1,-1,x) + LTEMP(step-1,1,x) +
                 	LTEMP(step-1,0,x-1) + LTEMP(step-1,0,x+1)) - 4.0*LTEMP(step-1,0,x)
            	);
            }
            if(coords[0] == (dims[0]-1)){
            	//can not use south border for y=my since my+1 is wrong
            	LTEMP(step, my,x) = LTEMP(step-1,my,x) + LMAT(my,x) * (
                	(2.0*LTEMP(step-1,my-1,x) +
                 	LTEMP(step-1,my,x-1) + LTEMP(step-1,my,x+1)) - 4.0*LTEMP(step-1,my,x)
            	);
            }else{
            	LTEMP(step,my,x) = LTEMP(step-1,my,x) + LMAT(my,x) * (
                	(LTEMP(step-1,my-1,x) + LTEMP(step-1,my+1,x) +
                 	LTEMP(step-1,my,x-1) + LTEMP(step-1,my,x+1)) - 4.0*LTEMP(step-1,my,x)
            	);
            }
			for ( int y=1; y<my; y++ )
				LTEMP(step,y,x) = LTEMP(step-1,y,x) + LMAT(y,x) * (
                	(LTEMP(step-1,y-1,x) + LTEMP(step-1,y+1,x) +
                 	LTEMP(step-1,y,x-1) + LTEMP(step-1,y,x+1)) - 4.0*LTEMP(step-1,y,x)
            	);
		}

	}


	//Update northborder if there is a neighboor in north
	if(coords[0] != 0){
		for(int y=startPos;y<0;y++){
			if(coords[1] == 0){
				//can not use westborder for x=0
				LTEMP(step,y,0) = LTEMP(step-1,y,0) + LMAT(y,0) * (
                	(LTEMP(step-1,y-1,0) + LTEMP(step-1,y+1,0) +
                 	2.0*LTEMP(step-1,y,1)) - 4.0*LTEMP(step-1,y,0)
            	);
			}else{
				LTEMP(step,y,0) = LTEMP(step-1,y,0) + LMAT(y,0) * (
                	(LTEMP(step-1,y-1,0) + LTEMP(step-1,y+1,0) +
                 	LTEMP(step-1,y,-1) + LTEMP(step-1,y,1)) - 4.0*LTEMP(step-1,y,0)
            	);
			}
			if(coords[1] == (dims[1]-1)){
				//can not use eastborder for x=mx
				LTEMP(step,y,mx) = LTEMP(step-1,y,mx) + LMAT(y,mx) * (
                	(LTEMP(step-1,y-1,mx) + LTEMP(step-1,y+1,mx) +
                 	2.0*LTEMP(step-1,y,mx-1))  - 4.0*LTEMP(step-1,y,mx)
            	);
			}else{
				LTEMP(step,y,mx) = LTEMP(step-1,y,mx) + LMAT(y,mx) * (
                	(LTEMP(step-1,y-1,mx) + LTEMP(step-1,y+1,mx) +
                 	LTEMP(step-1,y,mx-1) + LTEMP(step-1,y,mx+1)) - 4.0*LTEMP(step-1,y,mx)
            	);
			}
			for(int x=1;x<mx;x++){
				LTEMP(step,y,x) = LTEMP(step-1,y,x) + LMAT(y,x) * (
                	(LTEMP(step-1,y-1,x) + LTEMP(step-1,y+1,x) +
                 	LTEMP(step-1,y,x-1) + LTEMP(step-1,y,x+1)) - 4.0*LTEMP(step-1,y,x)
            	);
			}
		}
	}


	//Update southborder if there is a neighboor in south
	if(coords[0] != (dims[0]-1)){
		for(int y=(my-startPos);y>my; y--){
			if(coords[1] == 0){
				//can not use westborder for x=0
				LTEMP(step,y,0) = LTEMP(step-1,y,0) + LMAT(y,0) * (
                	(LTEMP(step-1,y-1,0) + LTEMP(step-1,y+1,0) +
                 	2.0*LTEMP(step-1,y,1)) - 4.0*LTEMP(step-1,y,0)
            	);
			}else{
				LTEMP(step,y,0) = LTEMP(step-1,y,0) + LMAT(y,0) * (
                	(LTEMP(step-1,y-1,0) + LTEMP(step-1,y+1,0) +
                 	LTEMP(step-1,y,-1) + LTEMP(step-1,y,1)) - 4.0*LTEMP(step-1,y,0)
            	);
			}
			if(coords[1] == (dims[1]-1)){
				//can not use eastborder for x=mx
				LTEMP(step,y,mx) = LTEMP(step-1,y,mx) + LMAT(y,mx) * (
                	(LTEMP(step-1,y-1,mx) + LTEMP(step-1,y+1,mx) +
                 	2.0*LTEMP(step-1,y,mx-1)) - 4.0*LTEMP(step-1,y,mx)
            	);
			}else{
				LTEMP(step,y,mx) = LTEMP(step-1,y,mx) + LMAT(y,mx) * (
                	(LTEMP(step-1,y-1,mx) + LTEMP(step-1,y+1,mx) +
                 	LTEMP(step-1,y,mx-1) + LTEMP(step-1,y,mx+1)) - 4.0*LTEMP(step-1,y,mx)
            	);
			}
			for(int x=1;x<mx;x++){
				LTEMP(step,y,x) = LTEMP(step-1,y,x) + LMAT(y,x) * (
                	(LTEMP(step-1,y-1,x) + LTEMP(step-1,y+1,x) +
                 	LTEMP(step-1,y,x-1) + LTEMP(step-1,y,x+1)) - 4.0*LTEMP(step-1,y,x)
            	);
			}

		}
	}
}

void
commit_vector_types ( void )
{
    MPI_Type_vector ( heightSIZE/dims[0], local_dims[1], dims[1]*local_dims[1],
        MPI_FLOAT, &global_area
    );
    MPI_Type_vector ( heightSIZE/dims[0], local_dims[1], local_dims[1]+2*border ,
        MPI_FLOAT, &local_area
    );
    MPI_Type_commit ( &local_area );
    MPI_Type_commit ( &global_area );

    /* Commit the types for the border exchange */
    MPI_Type_vector ( border , local_dims[1]+2*border, local_dims[1]+2*border ,
        MPI_FLOAT, &border_row
    );
    MPI_Type_vector ( local_dims[0], border , local_dims[1]+2*border ,
        MPI_FLOAT, &border_col
    );
    MPI_Type_commit ( &border_row );
    MPI_Type_commit ( &border_col );
}


void
border_exchange ( int step )
{
    /* east -> me -> west */
    MPI_Sendrecv (
        &LTEMP(step,0,0), 1, border_col, west, 0,
        &LTEMP(step,0,local_dims[1]), 1, border_col, east, 0,
        cart, MPI_STATUS_IGNORE
    );
    /* west -> me -> east */
    MPI_Sendrecv (
        &LTEMP(step,0,local_dims[1]-border ), 1, border_col, east, 0,
        &LTEMP(step,0,-border ), 1, border_col, west, 0,
        cart, MPI_STATUS_IGNORE
    );
    /* south -> me -> north */
    MPI_Sendrecv (
        &LTEMP(step,0,-border ), 1, border_row, north, 0,
        &LTEMP(step,local_dims[0],-border ), 1, border_row, south, 0,
        cart, MPI_STATUS_IGNORE
    );
    /* north -> me -> south */
    MPI_Sendrecv (
        &LTEMP(step,local_dims[0]-border ,-border ), 1, border_row, south, 0,
        &LTEMP(step,-border,-border), 1, border_row, north, 0,
        cart, MPI_STATUS_IGNORE
    );
}
void printTemp(int step){
    printf("rank %d local dims y %d x%d local_origin y%d x%d\n", rank,local_dims[0],local_dims[1], local_origin[0], local_origin[1]);

    for ( int y=0; y<local_dims[0]+(2*border); y++ ){
        for ( int x=0; x<local_dims[1]+(2*border); x++ )
        {
                printf(" %4.2f ",local_temp[(step%2)][y*(local_dims[1]+(2*border)) + x]);
        }
        printf("\n");
    }
}
void printMat(){
    printf("rank %d local dims y %d x%d local_origin y%d x%d\n", rank,local_dims[0],local_dims[1], local_origin[0], local_origin[1]);

    for ( int y=0; y<local_dims[0]+(2*border); y++ ){
        for ( int x=0; x<local_dims[1]+(2*border); x++ )
        {
                printf(" %4.2f ",local_material[y*(local_dims[1]+(2*border)) + x]);
        }
        printf("\n");
    }
}

void logTime( void){
        char* filename = "log.txt";
        FILE *out = fopen ( filename, "a" );
        if (out == NULL)
        {
            printf("Error opening file!\n");
            exit(1);
        }
        fprintf ( out, "%f;%d;%d;%d;%d;%d;%d;%d;%d;runtime systemSIZE border NSTEPS CUTOFF size y x write to file \n", end-start, systemSIZE, border,  NSTEPS, CUTOFF, size, dims[0], dims[1], WRITETOFILE);
        fclose ( out );

}

void printStatus(int step){
	int tilfeldig = rank;
    if(rank == 0){
        //printMat();
        //printf("****************************\n");
        printTemp(step);
        MPI_Send(&tilfeldig,1,MPI_INT, 1, 0, MPI_COMM_WORLD);
    }else{
        MPI_Recv(&tilfeldig,1,MPI_INT, rank-1,0,MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        //printMat();
        //printf("****************************\n");
        printTemp(step);
        if(rank != size-1){
            MPI_Send(&tilfeldig,1,MPI_INT, rank+1, 0, MPI_COMM_WORLD);
        }
    }

}

void printLargeTemperature(){
	if(rank == 0){
    	for ( int y=0; y<widthSIZE; y++ ){
        	for ( int x=0; x<heightSIZE; x++ )
        	{
                printf(" %3.1f ",temperature[y*widthSIZE + x]);
        	}
        	printf("\n");
    	}
    }
}

int
main ( int argc, char **argv )
{
    MPI_Init ( &argc, &argv );
    MPI_Comm_size ( MPI_COMM_WORLD, &size );
    MPI_Comm_rank ( MPI_COMM_WORLD, &rank );

    if(argc > 1){
    	if((atoi(argv[1])%256)==0){
    		systemSIZE = atoi(argv[1]);
    		widthSIZE = systemSIZE;
    		heightSIZE = systemSIZE;
    	}else{
    		if(rank == 0)
    			printf("Size error %s is not diviable by 256\n", argv[1]);
    	}
        if(atoi(argv[2])<10){
        		border = atoi(argv[2]);
        }else{
        	if(rank == 0)
        		printf("%s is too large as a border \n", argv[2]);
        }
        //printf("%d %d rank %d \n", systemSIZE, border, rank);

    }
    /* Start timing */
    MPI_Barrier(MPI_COMM_WORLD);
    start = MPI_Wtime();


    MPI_Dims_create ( size, 2, dims );
    MPI_Cart_create ( MPI_COMM_WORLD, 2, dims, periods, 0, &cart );
    MPI_Cart_coords ( cart, rank, 2, coords );
    MPI_Cart_shift ( cart, 0, 1, &north, &south );
    MPI_Cart_shift ( cart, 1, 1, &west, &east );

    local_dims[0] = local_realdims[0] = systemSIZE / dims[0];
    local_dims[1] = local_realdims[1] = systemSIZE / dims[1];

    //Pad the matrixes so they are diviable by the dimensions
   	if((local_dims[0]*dims[0])!= systemSIZE){
    	//padding to make it bigger
    	local_realdims[0] += 1;
		local_dims[0] += 1;

    	heightSIZE = local_dims[0]*dims[0];

    	padding[0] = heightSIZE - systemSIZE;

    	if(coords[0] == (dims[0]-1)){
    		//threads at the bottom should not compute the padding
    		local_realdims[0] -= padding[0];
    		if(padding[0]> local_dims[0]){
    			local_realdims[0] = 0;
    		}
    	}
    	if(padding[0]> local_dims[0]){
    		//fix padding bigger than local_dims[0] problem
    		if((padding[0]-(local_dims[0]*(dims[0]-coords[0]))) > 0){
    			local_realdims[0] -= padding[0]-(local_dims[0]*(dims[0]-coords[0]));
    			if(local_realdims[0]<0)
    				local_realdims[0] = 0;
    		}
    	}
    	//printf("rank %d is %d x %d with real %d x %d and padding %d %dheighsize %d widthSIZE%d\n", rank, local_dims[0], local_dims[1], local_realdims[0], local_realdims[1], padding[0], padding[1], heightSIZE, widthSIZE);
    }
    if((local_dims[1]*dims[1])!= systemSIZE){
    	//padding to make it bigger
    	local_realdims[1] += 1;
		local_dims[1] += 1;

    	widthSIZE = local_dims[1]*dims[1];

    	padding[1] = widthSIZE - systemSIZE;

    	if(coords[1] == (dims[1]-1)){
    		//rightmost threads should not compute the padding
    		local_realdims[1] -= padding[1];
    	}
        if(padding[1]> local_dims[1]){
    		//fix padding bigger than local_dims[0] problem
    		if((padding[1]-(local_dims[1]*(dims[1]-coords[1]))) > 0){
    			local_realdims[1] -= padding[1]-(local_dims[1]*(dims[1]-coords[1]));
    			if(local_realdims[1]<0)
    				local_realdims[1] = 0;
    		}
    	}
    	//printf("rank %d is %d x %d with real %d x %d and padding %d %dwidthsize %d height %d \n", rank, local_dims[0], local_dims[1], local_realdims[0], local_realdims[1], padding[0], padding[1], widthSIZE, heightSIZE);
    }

    local_origin[0] = coords[0]*local_dims[0];
    local_origin[1] = coords[1]*local_dims[1];

    size_t lsize_full = (local_dims[0]+2*border)*(local_dims[1]+2*border);
    local_material = malloc ( lsize_full * sizeof(float) );
    local_temp[0] = malloc ( lsize_full * sizeof(float) );
    local_temp[1] = malloc ( lsize_full * sizeof(float) );

    if ( rank == 0 ){
    	temperature = calloc(widthSIZE*heightSIZE, sizeof(float));
    }

    commit_vector_types (); //Commit Vector types for borderexchange
    configure_geometry(); //Set up the LMAT and LTEMP

    /* Main integration loop: NSTEPS iterations, impose external heat
     * until CUTOFF iterations have passed
     */
     MPI_Barrier(MPI_COMM_WORLD);
    /* Imposed temperature from outside */
    for ( int step=0; step<NSTEPS; step++ )
    {
        if ( step < CUTOFF )
            external_heat ( step );

        if((step%border)==0){
        	//Exchange the border every borderth step
        	border_exchange ( step );
        }else{
        	//Need to be updated so the FTCS solver can use the row or column closest to itself
        	update_border( step );
        }
        ftcs_solver ( step );
        boundaries ( step );

        if((step % SNAPSHOT) == 0 )
        {
            char filename[15];
            sprintf ( filename, "heat.dat" );
            collect_area ( step, filename );
        }

    }

    /* End timing */
    MPI_Barrier(MPI_COMM_WORLD);
    end = MPI_Wtime();

    /*
    //Find processorname
    int name_len;
    char processor_name[20];
    MPI_Get_processor_name(processor_name, &name_len);
    printf("rank %d har processor %s\n", rank, processor_name );
	*/

    if ( rank == 0 )
        free (temperature);
    free (local_material), free (local_temp[0]), free (local_temp[1]);
    MPI_Finalize();

    /* Print out timing */
    if (rank == 0) {
        printf("RUNTIME %4.6f SIZE %d border %d NSTEPS %d CUTOFF%d THREADS %d in y %d x %d\n", end-start, systemSIZE, border, NSTEPS, CUTOFF, size, dims[0], dims[1]);
	    logTime();
    }

    exit ( EXIT_SUCCESS );
}


void
external_heat ( int step )
{
    /* Imposed temperature from outside */
    for ( int y=(systemSIZE/2)-(systemSIZE/16); y<=(systemSIZE/2)+(systemSIZE/16); y++ )
        for ( int x=(systemSIZE/4); x<=(3*systemSIZE/4); x++ )
        {
            if ( BORBOX(y,x) )
                LTEMP ( step, y-local_origin[0], x-local_origin[1] ) = 100.0;
        }
}


void
configure_geometry ( void )
{
    /* Initialization: fill the pool with mercury */
    for ( int y=-border; y<(local_dims[0]+border); y++ )
    {
        for ( int x=-border; x<(local_dims[1]+border); x++ )
        {
            LMAT(y,x) =  MERCURY * (dt/(h*h));
            LTEMP(1,y,x) = LTEMP(0,y,x) = 20.0;
        }
    }

    /* Set up the two blocks of copper and tin */
    for ( int y=(systemSIZE/8); y<(3*systemSIZE/8); y++ )
        for ( int x=(systemSIZE/8); x<(systemSIZE/2)-(systemSIZE/8); x++ )
        {
            if ( BORBOX(y,x) )
            {
                LMAT(y-local_origin[0],x-local_origin[1]) =
                    COPPER * (dt/(h*h));
                LTEMP(0,y-local_origin[0],x-local_origin[1]) = 60.0;
            }
            if ( BORBOX(y,systemSIZE-x) )
            {
                LMAT(y-local_origin[0],(systemSIZE-x)-local_origin[1]) =
                    TIN * (dt/(h*h));
                LTEMP(0,y-local_origin[0],(systemSIZE-x)-local_origin[1]) = 60.0;
            }

        }
    /* Set up the heating element in the middle */
    for ( int y=(systemSIZE/2)-(systemSIZE/16); y<=(systemSIZE/2)+(systemSIZE/16); y++ )
        for ( int x=(systemSIZE/4); x<=(3*systemSIZE/4); x++ )
        {
            if ( BORBOX(y,x) )
                LMAT(y-local_origin[0],x-local_origin[1]) =
                    ALUMINIUM * (dt/(h*h));
        }
}


void
collect_area ( int step, char *filename )
{
	//TODO:fiks avrundsfeil her
    MPI_Request req;
    MPI_Isend ( &LTEMP((step%SNAPSHOT),0,0), 1, local_area, 0, 0, cart, &req );
    //TODO: send siste del av side eller bunn
    if ( rank == 0 )
    {
        int co[2];
        for ( int r=0; r<size; r++ )
        {
            MPI_Cart_coords ( cart, r, 2, co );
            MPI_Recv (
                &TEMP( co[0]*local_dims[0], co[1]*local_dims[1] ),
                    1, global_area, r, 0, cart, MPI_STATUS_IGNORE
            );
        }
        if(WRITETOFILE){
        	FILE *out = fopen ( filename, "w" );
        	write_matrix ( out, temperature );
        	fclose ( out );
        	printf ( "Snapshot at step %d\n", step );
        }
    }
    MPI_Wait ( &req, MPI_STATUS_IGNORE );
}


void
write_matrix ( FILE *out, float *data )
{
    float size = (float) systemSIZE;
    fwrite ( &size, sizeof(float), 1, out );
    for ( float x=0; x<systemSIZE; x+=1.0 )
        fwrite ( &x, sizeof(float), 1, out );
    for ( int y=0; y<systemSIZE; y++ )
    {
        float len = (float) y;
        fwrite ( &len, sizeof(float), 1, out );
        fwrite ( &data[y*widthSIZE], sizeof(float), systemSIZE, out );
    }
}
